/*
 * $Id: wmLifeMeter.c,v 1.13 2004/05/24 22:47:31 shaper Exp $
 *
 * Displays a dynamically-updating progress bar and numerical digits
 * representing your progress thus far in life.
 *
 * 07/11/03 - Walter Korman (shaper@waywardgeeks.org)
 */

/*
    wmLifeMeter - humorous display of life elapsed.
    Copyright (C) 2003 Walter Korman

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
    USA
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "xwin.h"

/* defines */
#define VERSION "0.4"

#define DEBUG (0)

#define DELAY (10000L)
#define SECONDS_PER_DAY (60 * 60 * 24)
#define DAYS_PER_YEAR (365)
#define EPOCH_START_YEAR (1970)

#define MIN_DEATH_AGE (1)
#define MAX_DEATH_AGE (136)
#define MIN_YEAR (1900)

#define PIE_THETA0 ((360 / 4) * 64)
#define CIRCLE_THETA (360 * 64)

#define DM_PIE (0)
#define DM_BAR (1)

#define DEFAULT_BGCOLOR "white"
#define DEFAULT_FGCOLOR "yellow"
#define DEFAULT_TXCOLOR "black"
#define DEFAULT_BGPIECOLOR "blue"

#define DEFAULT_DISPLAYMODE DM_PIE

/* global variables */
int bday, bmonth, byear;
int deathAge;
double lifeElapsed;
int isOffscreenDirty = 1;
char* fgcolor = DEFAULT_FGCOLOR;
char* bgcolor = DEFAULT_BGCOLOR;
char* txcolor = DEFAULT_TXCOLOR;
char* bgpiecolor = DEFAULT_BGPIECOLOR;
unsigned long fgcolorpix, bgcolorpix, txcolorpix, bgpiecolorpix, framecolorpix;
int dmode = DEFAULT_DISPLAYMODE;
char* name = NULL;

/* function prototypes */
double getLifeElapsed (int bmonth, int bday, int byear, int deathAge);
time_t birthdayToEpochSeconds (int month, int day, int year);
void handleButtonPress (XButtonEvent* xev);
void parseCommands (int argc, char* argv[]);
void printUsage ();
void redrawWindow ();
void updateOffscreen ();

/*
 * The main business.
 */
int
main (int argc, char *argv[])
{
    double newLifeElapsed;
    XEvent event;

    /* parse command-line arguments */
    parseCommands(argc, argv);

    /* determine our starting elapsed lifetime */
    lifeElapsed = getLifeElapsed(bmonth, bday, byear, deathAge);

    /* create and initialize the application window */
    xw_initXwindow(argc, argv);
    xw_openXwindow(argc, argv, bgcolor, fgcolor);

    /* determine the color mappings */
    fgcolorpix = xw_getColor(fgcolor);
    bgcolorpix = xw_getColor(bgcolor);
    txcolorpix = xw_getColor(txcolor);
    bgpiecolorpix = xw_getColor(bgpiecolor);
    framecolorpix = xw_getColor(FRAME_COLOR);

    while (1) {
        /* process any pending X events */
        while(XPending(display)){
            XNextEvent(display, &event);

            switch(event.type){
            case Expose:
                redrawWindow();
                break;

            case ButtonPress:
                handleButtonPress(&event.xbutton);
                break;

            case ButtonRelease:
                break;
            }
        }

        /* redraw the window */
        redrawWindow();

        /* wait for a bit */
        usleep(DELAY);

        /* check our elapsed life and update if appropriate */
        newLifeElapsed = getLifeElapsed(bmonth, bday, byear, deathAge);
        if (newLifeElapsed != lifeElapsed) {
            lifeElapsed = newLifeElapsed;
            isOffscreenDirty = 1;
        }
    }

    return 0;
}

/*
 * Parses command-line arguments.
 */
void
parseCommands (int argc, char* argv[])
{
    time_t nowSecs;
    struct tm* now;
    int curYear;
    int cidx;

    /* consider all optional arguments */
    for (cidx = 1; cidx < argc; cidx++) {
        char* arg = argv[cidx];

#if DEBUG
        fprintf(stderr, "considering [arg=%s, cidx=%d].\n", arg, cidx);
#endif

        if (!strcmp(arg, "-version")) {
            fprintf(stdout, "wmLifemeter version %s\n", VERSION);
            exit(EXIT_SUCCESS);

        } else if (!strcmp(arg, "-bgcolor") && cidx < (argc - 1)) {
            bgcolor = argv[++cidx];

        } else if (!strcmp(arg, "-fgcolor") && cidx < (argc - 1)) {
            fgcolor = argv[++cidx];

        } else if (!strcmp(arg, "-txcolor") && cidx < (argc - 1)) {
            txcolor = argv[++cidx];

        } else if (!strcmp(arg, "-bgpiecolor") && cidx < (argc - 1)) {
            bgpiecolor = argv[++cidx];

        } else if (!strcmp(arg, "-dmode") && cidx < (argc - 1)) {
            cidx++;
            if (!strcmp(argv[cidx], "bar")) {
                dmode = DM_BAR;

            } else if (!strcmp(argv[cidx], "pie")) {
                dmode = DM_PIE;

            } else {
                fprintf(stderr, "Unknown display mode: %s\n", argv[cidx]);
                printUsage();
                exit(EXIT_FAILURE);
            }

        } else if (!strcmp(arg, "-name") && cidx < (argc - 1)) {
            name = argv[++cidx];

        } else {
            break;
        }
    }

    /* make sure we've still got the required arguments */
    if ((argc - cidx) < 2) {
        printUsage();
        exit(EXIT_FAILURE);
    }

    /* read the birthday */
    int count = sscanf(argv[cidx++], "%d/%d/%d", &bmonth, &bday, &byear);
    if (count != 3) {
        printUsage();
        exit(EXIT_FAILURE);
    }

    /* read the age at death */
    deathAge = atoi(argv[cidx++]);

#if DEBUG
    fprintf(stderr, "month=%d, day=%d, year=%d, deathAge=%d\n",
            bmonth, bday, byear, deathAge);
#endif

    /* sanity-check parameters */

    if (deathAge < MIN_DEATH_AGE || deathAge > MAX_DEATH_AGE) {
        fprintf(stderr, "Life expectancy must be between %d and %d.\n",
                MIN_DEATH_AGE, MAX_DEATH_AGE);
        exit(EXIT_FAILURE);
    }

    /* get the current year */
    nowSecs = time(NULL);
    now = localtime(&nowSecs);
    curYear = now->tm_year + 1900;

    if ((bmonth < 1 || bmonth > 12) || (bday < 1 || bday > 31) ||
        (byear < MIN_YEAR || byear > curYear)) {
        fprintf(stderr, "Birthday must be a valid date between %d and now.\n",
                MIN_YEAR);
        exit(EXIT_FAILURE);
    }
}

/*
 * Returns the percent life elapsed for someone born on the specified date
 * and expected to die at the specified age in years.
 */
double
getLifeElapsed (int bmonth, int bday, int byear, int deathAge)
{
    unsigned long lifeSecs, bornSecs, todaySecs, ageSecs, epochDeltaSecs;
    long epochDeltaYears;
    double lifeElapsed;

    /* determine overall life in seconds */
    lifeSecs = deathAge * DAYS_PER_YEAR * SECONDS_PER_DAY;

    /* determine the epoch delta to allow calculating for folks whose
     * birthdays are before the epoch */
    if ((epochDeltaYears = EPOCH_START_YEAR - byear) > 0) {
        epochDeltaSecs = (epochDeltaYears * DAYS_PER_YEAR * SECONDS_PER_DAY);

    } else {
        epochDeltaYears = 0;
        epochDeltaSecs = 0;
    }

    /* determine the seconds elapsed since the epoch at birth */
    bornSecs = birthdayToEpochSeconds(bmonth, bday, byear + epochDeltaYears);

    /* determine the person's age in seconds */
    todaySecs = time(NULL) + epochDeltaSecs;
    ageSecs = todaySecs - bornSecs;

    /* determine the percent life elapsed */
    lifeElapsed = (lifeSecs <= 0) ? 0.0 : (ageSecs / (double)lifeSecs);

#if DEBUG
    /* log some useful information */
    fprintf(stderr, "month=%d, day=%d, year=%d\n", bmonth, bday, byear);
    fprintf(stderr, "deathAge=%d\n", deathAge);
    fprintf(stderr, "edyears=%ld, edsecs=%lu\n", epochDeltaYears,
            epochDeltaSecs);
    fprintf(stderr, "ageSecs=%lu, lifeSecs=%lu, lifeElapsed=%0.9f\n",
            ageSecs, lifeSecs, lifeElapsed);
#endif

    /* constrain life elapsed to a maximum of 100% */
    return (lifeElapsed > 1.0) ? 1.0 : lifeElapsed;
}

/*
 * Prints the usage string for the application.
 */
void
printUsage ()
{
    printf("Usage: wmLifemeter <birthday: MM/DD/YYYY> <lifeInYears: 70>\n"
           "Options:\n"
           "  -version    : print version information\n"
           "  -fgcolor    : display foreground color ('red', 'blue' ...)\n"
           "  -bgcolor    : display background color\n"
           "  -txcolor    : display text color\n"
           "  -bgpiecolor : display background pie chart color\n"
           "  -dmode      : display mode ('bar', 'pie')\n"
           "  -name       : doomed person's name\n");

    /* -font-family 'terminal' */
    /* -font-size 14 */
    /* -name shaper */
}

/*
 * Returns the number of seconds from the start of the epoch to the
 * specified date.  Returns 0 for all pre-1970 dates.
 */
time_t
birthdayToEpochSeconds (int month, int day, int year)
{
    struct tm tm;

    memset(&tm, 0, sizeof(struct tm));
    tm.tm_mday = day;
    tm.tm_mon = (month - 1);
    tm.tm_year = (year - 1900);
    return mktime(&tm);
}

/*
 * Redraw the main and icon window.
 */
void
redrawWindow ()
{
    if (NormalGC == NULL || OffscreenGC == NULL) {
        return;
    }

    /* note that we've dealt with our exposure */
    xw_flushWindowExposure(iconwin);

    /* update the offscreen if appropriate */
    if (isOffscreenDirty) {
        updateOffscreen();
        isOffscreenDirty = 0;
    }

    /* copy the offscreen pixmap to the window */
    XCopyArea(display, offpixmap, iconwin, NormalGC, 0, 0,
              DISPLAY_WIDTH, DISPLAY_HEIGHT, 0, 0);

    /* note that we've dealt with our exposure */
    xw_flushWindowExposure(win);

    /* copy the offscreen pixmap to the window */
    XCopyArea(display, offpixmap, win, NormalGC, 0, 0, DISPLAY_WIDTH,
              DISPLAY_HEIGHT, 0, 0);
}

/*
 * Renders the enter image comprising the current display of the person's
 * elapsed life to the offscreen bitmap.
 */
void
updateOffscreen ()
{
    char lifestr[255];
    int rwid = (int) (DISPLAY_WIDTH * lifeElapsed);
    int swid, shei, x, y;
    int lifeAngle;
    double lifePercent;
    XGCValues gcv;

#if DEBUG
        fprintf(stderr, "updateOffscreen\n");
#endif

    /* draw the background */
    xw_setColor(display, OffscreenGC, bgcolorpix);
    XFillRectangle(display, offpixmap, OffscreenGC, 0, 0, DISPLAY_WIDTH - 1,
                   DISPLAY_HEIGHT - 1);

    /* draw the life progress display */
    switch (dmode) {
    case DM_BAR:
        /* draw the life progress bar */
        xw_setColor(display, OffscreenGC, fgcolorpix);
        XFillRectangle(display, offpixmap, OffscreenGC, 0, 0, rwid,
                       DISPLAY_HEIGHT);
        break;

    case DM_PIE:
        gcv.arc_mode = ArcPieSlice;
        XChangeGC(display, OffscreenGC, GCArcMode, &gcv);

        /* draw the background circle */
        xw_setColor(display, OffscreenGC, bgpiecolorpix);
        XFillArc(display, offpixmap, OffscreenGC, 0, 0,
                 DISPLAY_WIDTH - 1, DISPLAY_HEIGHT - 1,
                 PIE_THETA0, CIRCLE_THETA);

        /* draw the main pie chart progress */
        xw_setColor(display, OffscreenGC, fgcolorpix);
        lifeAngle = lifeElapsed * (CIRCLE_THETA);
        XFillArc(display, offpixmap, OffscreenGC, 0, 0,
                 DISPLAY_WIDTH - 1, DISPLAY_HEIGHT - 1,
                 PIE_THETA0, -lifeAngle);

        /* frame the edge of the chart */
        xw_setColor(display, OffscreenGC, framecolorpix);
        XDrawArc(display, offpixmap, OffscreenGC, 0, 0,
                 DISPLAY_WIDTH - 1, DISPLAY_HEIGHT - 1,
                 PIE_THETA0, CIRCLE_THETA);
        break;
    }

    /* draw the frame around all of our business  */
    xw_setColor(display, OffscreenGC, framecolorpix);
    XDrawRectangle(display, offpixmap, OffscreenGC, 0, 0, DISPLAY_WIDTH - 1,
                   DISPLAY_HEIGHT - 1);

    /* generate the elapsed lifetime strings */
    lifePercent = lifeElapsed * 100.0f;
    sprintf(lifestr, "%0.7f", lifePercent);

    /* determine centered positioning of life elapsed percentage text */
    XSetFont(display, OffscreenGC, displayFont);
    xw_getStringDimensions(display, OffscreenGC, lifestr, &swid, &shei);

    x = (DISPLAY_WIDTH - swid) / 2;
    y = ((DISPLAY_HEIGHT - shei) / 2) + shei;

    /* draw the percent life elapsed */
    xw_setColor(display, OffscreenGC, txcolorpix);
    XDrawString(display, offpixmap, OffscreenGC, x, y, lifestr,
                strlen(lifestr));

    if (name != NULL) {
        /* draw the name */
        xw_getStringDimensions(display, OffscreenGC, name, &swid, &shei);

        x = (DISPLAY_WIDTH - swid) / 2;
        y = (DISPLAY_BUFFER + shei);

        XDrawString(display, offpixmap, OffscreenGC, x, y, name, strlen(name));
    }
}

/*
 * Handles a button press on the window.
 */
void
handleButtonPress (XButtonEvent *xev)
{
    fprintf(stderr, "button press.\n");

    dmode = (dmode == DM_PIE) ? DM_BAR : DM_PIE;
    isOffscreenDirty = 1;
}
